###############################################################################
# Exprimentation : gestion des valeurs manquantes lors du dploiement 
# d'un modle issu de la rgression logistique
#
# La base est paramtrable : il faut que la variable  prdire s'appelle
# "classe" et soit situe en dernire colonne. Toutes les variables 
# prdictives doivent tre numriques.
###############################################################################

rm(list=ls())

#partition des donnes en apprentissage et test
#rpartiton par dfaut (apprentissage = 70%, test = 30 %)
partition <- function(donnees,prc=0.7){
	#effectifs
	n <- nrow(donnees)
	n.train <- trunc(n*prc)
	n.test <- n - n.train
	#index
	alea <- runif(n)
	index <- rank(alea,ties.method="random")
	#partition
	donnees.train <- subset(donnees,subset=(index <= n.train))
	donnees.test <- subset(donnees,subset=(index > n.train))
	#retour
	return(list(train=donnees.train,test=donnees.test))
}

#insre k NA au hasard dans une ligne
insert.na.row <- function(ligne,k){
	#nombre de colonnes dans la ligne
	J <- length(ligne)
	#index des colonnes  modifier
	index <- rank(runif(J))[1:k]
	#inserer le code NA
	y <- ligne
	y[index] <- NA
	return(y)	
}

#insre k NA par ligne dans un data.frame
#on considre que toutes les colonnes sont numriques
insert.na.data.frame <- function(donnees,k){
	#transformer les donnes en matrice
	#attention, il faut que toutes les colonnes soient numriques
	#dans le cas contraire, nous avons une matrice de chanes de caractres
	matrice <- as.matrix(donnees)
	#appliquer la modification - traitement par ligne
	new.matrice <- apply(matrice,1,insert.na.row,k)
	#transposer la matrice
	new.matrice <- t(new.matrice)
	#transformer en data.frame
	new.donnees <- as.data.frame(new.matrice)
	return(new.donnees)
}

#modele de prdiction par la rgression
#pour remplacer les valeurs d'une variable  partir des variables
#utilisation trs classique d'une boucle
modele.replace.by.regression <- function(donnees){
	modele.replace <- list()
	for (j in 1:ncol(donnees)){
		formule <- paste(names(donnees)[j]," ~ .",sep="")
		modele.replace[[j]] <- lm(as.formula(formule),data=donnees)
	}
	names(modele.replace) <- names(donnees)
	return(modele.replace)
}

#modele de prdiction par la rgression
#pour remplacer les valeurs d'une variable  partir des variables
#fonction pour appel de lapply
modele.replace.by.reg.for.lapply <- function(x,donnees){
	#rcuprer le numro de x dans le data.frame
	#cf. http://forums.cirad.fr/logiciel-R/viewtopic.php?p=15851&sid=60c7e747fa7c9778c915ed91443a6615
	numero <- as.numeric(gsub("\\D","", deparse(substitute(x)), perl=T))
	#rgression de x contre toutes les autres
	modele <- lm(x ~ ., data = subset(donnees,select = -numero))
	return(modele)
}

#remplacement des valeurs manquantes de x par la moyenne
replace.by.mean.one <- function(x, means){
	numero <- as.numeric(gsub("\\D","", deparse(substitute(x)), perl=T))
	y <- ifelse(is.na(x)==T, means[numero],x)
	return(y)
}

#remplacement des valeurs manquantes par la rgression
#on s'appuie sur les valeurs des colonnes o les NA ont t remplaces par les moyennes
replace.by.reg.one <- function(x, modeles, donnees.mean){
	numero <- as.numeric(gsub("\\D","", deparse(substitute(x)), perl=T))
	pred <- predict(modeles[[numero]],newdata=subset(donnees.mean,select=-numero))
	y <- ifelse(is.na(x) == T,pred,x)
	return(y)
}

#construction matrice de confusion  partir du modle
#new.dataset reprsente l'chantillon test
#model est le modle
pred_and_confusion_matrix <- function(new.dataset,model.glm){
	#probabilit prdit par le modle
	data.pred.prob <- predict(model.glm,newdata = new.dataset)
	#affectation
	data.pred.class <- ifelse(data.pred.prob > 0.5,"B","A")
	data.pred.class <- as.factor(data.pred.class)
	#matrice de confusion (classe observe vs. prdite)
	mc <- table(new.dataset$classe,data.pred.class)
	#taux d'erreur
	err.rate <- 1.0 - (mc[1,1]+mc[2,2])/sum(mc)
	return(err.rate)
}

#monter une exprimentation
experiments <- function(donnees.test, modele.glm, k){
	#rcuprer les colonnes sauf la classe
	donnees.numeric <- subset(donnees.test,select= -classe)
	
	#insrer les donnes manquantes
	donnees.na <- insert.na.data.frame(donnees.numeric,k)
	
	#remplacement des NA par la moyenne
	donnees.test.mean <- as.data.frame(lapply(donnees.na,replace.by.mean.one,donnees.numeric.mean))
	donnees.for.test <- cbind(donnees.test.mean,donnees.test$classe)
	names(donnees.for.test) <- names(donnees.test)
	e1 <- pred_and_confusion_matrix(donnees.for.test,modele.glm)
	
	#remplacement des NA par la rgression
	donnees.test.reg <- as.data.frame(lapply(donnees.na,replace.by.reg.one,modeles.replace.by.reg,donnees.test.mean))
	donnees.for.test <- cbind(donnees.test.reg,donnees.test$classe)
	names(donnees.for.test) <- names(donnees.test)
	e2 <- pred_and_confusion_matrix(donnees.for.test,modele.glm)
	
	#renvoyer le vecteur de 2 valeurs
	return(c(e1,e2))
}


#************************
#DEBUT DU PROG. PRINCIPAL
#PARAMETRES DE CALCUL
#************************

#nom du fichier  charger
NOM_FICHIER <- "breast.txt"

#proportion des observations en apprentissage
PROPORTION_TRAIN <- 0.50

#nombre max. de valeurs manquantes par ligne
M <- 7

#nombre de rptitions de chaque configuration
N <- 200


#*****************************************************************
#chargement des donnes, partition apprentissage/test
#construction du modle prdictif (rgression logistique avec glm)
#construction des modles de remplacement des valeurs manquantes
#*****************************************************************

#changement du rpertoire courante
setwd("D:/DataMining/Databases_for_mining/dataset_for_soft_dev_and_comparison/missing_data/deployment")
#chargement des donnes
donnees.all <- read.table(file=NOM_FICHIER,sep="\t",header=T,dec=".")

#initialiser le gnrateur de nombres alatoires
set.seed(100)

#partition en apprentissage-test (-> donnees$train et donnees$test)
donnees <- partition(donnees.all,prc=PROPORTION_TRAIN)

#construction du modele dfinitif sur les donnes d'apprentissage (donnees$train)
modele.glm <- glm(classe ~ ., data = donnees$train,family=binomial)

#mesurer le taux d'erreur de reference (donnes compltes : donnees$test)
err.ref <- pred_and_confusion_matrix(donnees$test,modele.glm)
print(paste("Erreur sur observations completes =",as.character(err.ref)))

#***********************************************
#construction des modles de remplacement des NA
#***********************************************
#travailler sur les variables prdictives
donnees.train.numeric <- subset(donnees$train,select = -classe)

#calculer le dterminant de la matrice de corrlation
#==> indique le degr de colinarit entre les variables explicatives
d <- det(cor(donnees.train.numeric))
print(paste("Determinant matrice de correlation =",as.character(d)))

#calculer les moyennes par colonne - chantillon d'apprentissage
donnees.numeric.mean <- sapply(donnees.train.numeric,mean)

#construire les modles de remplacement par la rgression
#modeles.replace.by.reg <- modele.replace.by.regression(donnees.train.numeric)
modeles.replace.by.reg <- lapply(donnees.train.numeric,modele.replace.by.reg.for.lapply,donnees.train.numeric)

#**********************************************************************
#lancer l'exprimentation
#pour un nombre de valeurs manquantes allant de 1  M pour chaque ligne
#de l'chantillon test
#>> rpeter N fois l'opration
#**********************************************************************

#pour collecter les rsultats
resultats <- c()
#exprimentation (s'il y a >> j << valeur manquante dans chaque ligne) 
for (k in 1:M){
	erreurs <- replicate(N,experiments(donnees$test,modele.glm,k))
	moyennes <- apply(erreurs,1,mean)
	resultats <- rbind(resultats,moyennes)
}
#rajouter les rsultats pour donnes compltes
resultats <- rbind(c(err.ref,err.ref),resultats)
rownames(resultats) <- 0:M
print(resultats)
#graphique
plot(0:M,resultats[,1],type="b",col="black",xlim=c(0,M),ylim=c(min(resultats),max(resultats)),xlab="#Missing values",ylab="Error rate")
points(0:M,resultats[,2],type="b",col="blue")

